; $Id: cfram.asm,v 1.3 2003/10/23 05:59:49 hpa Exp $
;
; Firmware for ABC80 CompactFlash controller
;
; This needs a CPU mostly because ABC wants 256-byte sectors, and
; CompactFlash has 512-byte sectors...
;
; Total memory: 2K
;    0-511  -> controller code and data
;  512-1023 -> space for one CF sector
; 1024-2047 -> space for four ABC sectors (RAMBUFs)
;
; FIX FIX FIX FIX FIX: Handle the case of removed card.  Although it
; seems odd to remove a hard disk, ABC-DOS should be able to handle it
; the same way it handles floppies.
;

reset				; Controller reset or C3
	im 1			; Interrupt = RST 38h
	ld hl,sysinfo
	ld de,sysinfo
	inc de
	ld bc,#13		; Size of sysinfo area - 1
	ld (hl),#0xDD		; Any value w/o E in high nybble
	ldir			; Clear status area

interrupt
	; Resume to begin command
	ld sp,#512			; Reset stack - we never actually RETI
	ld iy,sysinfo			; Restore IY

idle
	xor a
idle_err				; Errcode in A
	ei				; Allow interrupts

	; Prepare to receive command
	push af
	push iy				; Address for input
	pop hl
	ld a,l
	out (#12),a
	ld a,h
	out (#13),a
	ld a,#4
	out (#14),a
	pop af
	or #0x81			; Ready for command
	out (#8),a

wait_for_cmd
	in a,(#25)			; DMA status
	rrca
	jr nc,have_command		; Still waiting for bytes
	jr wait_for_cmd
	
	nop
	nop
	nop
	nop
	nop

interrupt_entry				; Address 0038h
	jr interrupt
	
have_command				; We have received an entire command sequence
	;
	; First, do common decoding operations
	;

	ld a,(iy+1)
	ld de,#0x100			; Controller buf ptr/4
	rlca				; 2 high bits of A -> 2 low bits of D
	rl d
	rlca
	rl d
	ld (abc_buf),de			; Save pointer to addressed controller buffer

	; Reset auxilliary status
	xor a
	out (#9),a
	
	;
	; READ SECTOR command
	;
	bit 0,(iy+0)
	jr z,not_read_sector
	
	call (load_buffer)		; Load the CF buffer with the addressed CF sector
	jr c,data_error

	; Copy relevant half of CF buffer to controller buffer
	call (bufsetup)
	ldir				; Copy into controller buffer
	jr not_read_sector

	nop
	nop
	nop
	nop
	nop
	nop
	nop
nmi_entry				; Address 0066h
	jp (nmi)
		
not_read_sector

	;
	; SECTOR TO HOST command
	;
	bit 1,(iy+0)			; SECTOR TO HOST
	jr z,not_sector_host
	ld e,#0x80			; Controller->CPU
	call (transfer)
not_sector_host

	;
	; SECTOR TO CONTROLLER command
	;
	bit 2,(iy+0)			; SECTOR TO CONTROLLER
	jr z,not_sector_controller
	ld e,#0				; CPU->controller
	call (transfer)
not_sector_controller

	;
	; WRITE SECTOR command
	;
	bit 3,(iy+0)
	jr z,not_write_sector
	
	; First, load the appropriate CF sector
	call (load_buffer)
	jr c,data_error

	; Copy controller buffer to relevant half of CF buffer
	call (bufsetup)
	ex de,hl			; Contoller buffer -> CF buffer
	ldir				; Copy into controller buffer
	
	; Finally, write back buffer to CF card
	call (write_buffer)
	jr c,data_error

not_write_sector

	;
	; All other commands ignored
	;
	xor a				; No error
	out (#9),a
	jp (idle)

	;
	; Place to jump when an error is detected; auxilliary error
	; code should be placed in register E
	;
data_error
	ld a,e				; Auxilliary status code
	out (#9),a
	ld a,#0x08			; ERROR
	jp (idle_err)

	;
	; Set up register for a CF buffer -> controller buffer
	; transfer for the currently addressed buffer and sector.
	; For the opposite direction transfer, do ex de,hl before
	; actually transferring data
	;
bufsetup
	ld hl,#0x0100			; CF buffer/2
	ld bc,#256			; 256 bytes
	ld a,(iy+3)			; Sector #
	rrca
	rl h				; HL -> proper half of CF buffer
	ld de,(abc_buf)			; Controller buffer
	ret
	
	;
	; Transfer a sector to or from the controller.  Register E
	; should be set to 0 for host->controller and 0x80 for
	; controller->host
	;
transfer
	xor a				; Controller buffer always aligned
	out (#12),a
	ld a,(iy+13)			; High byte of addressed controller buffer
	or e				; Direction bit
	out (#13),a
	xor a
	out (#15),a			; 256 bytes
	ld a,#0x01			; READY
	out (#8),a
tr_wait
	in a,(#25)			; DMA engine status
	rrca				; Bit 0 = DMA done
	ret nc
	jr tr_wait
	
;
; This routine converts the command contents to a CF sector number,
; and loads it if it is not currently loaded.
;
load_buffer
	; Initialize buffer addresses
	xor a
	ld (iy+9),a
	ld (iy+10),a
	ld (iy+11),#0xe0

	; Get track
	ld a,(iy+2)
	cp #238				; 238 tracks/volume
	ccf
	ld e,#0x10			; Seek error
	ret c				; Bad sector #
	srl a
	ld (iy+9),a

	; Get sector
	ld a,(iy+3)
	rr a				; Convert to CF sector + low byte of track
	ld (iy+8),a

	; Compute unit number -> third LBA byte
	ld a,(iy+1)			; K1 = unit # + more
	and #0x07			; Unit #
	jr z,lb_haveunit		; Unit 0 skips loop
	cp #0x04			; Only units 0-3 OK
	ccf
	ld e,#0x80			; Unit not ready error
	ret c				; Bad unit #
	; Each unit has 238 tracks, that is the maximum supported.
	; We multiply by 119 (238/2), this starts LBA bytes 2-3
	ld b,a
lb_getunit
	ld a,(iy+9)
	add a,#119
	ld (iy+9),a
	ld a,(iy+10)
	adc a,#0
	ld (iy+10),a
	djnz lb_getunit
lb_haveunit

	; Now cf_need contains the LBA sector (including upper 4 control bits)
	; that we need; cf_sect contains what's currently loaded in the CF
	; sector buffer.
	ld a,(iy+4)
	cp (iy+8)
	jr nz,lb_nomatch
	ld a,(iy+5)
	cp (iy+9)
	jr nz,lb_nomatch
	ld a,(iy+6)
	cp (iy+10)
	jr nz,lb_nomatch
	ld a,(iy+7)
	cp (iy+11)
	ret z				; If this matched, we're DONE...
lb_nomatch
	; No such luck.  Need to actually load the sector.
	ld a,#1
	out (#2),a			; Sector count
	ld a,(iy+8)
	out (#3),a
	ld a,(iy+9)
	out (#4),a
	ld a,(iy+10)
	out (#5),a
	ld a,(iy+11)
	out (#6),a
	ld a,#0x20			; Read sector
	di				; No interrupts that will cause CF card to be confused
	out (#7),a

	ld bc,#0x0000			; Port 0, 256 bytes
	ld hl,#512			; CF card buffer
	
lb_waitdrq
	in a,(#7)			; CF status
	bit 0,a
	jr nz,lb_error			; Error!
	bit 3,a				; Wait for DRQ
	jr z,lb_waitdrq

	ld (iy+7),#0			; No valid sector # in buffer ATM

	inir				; Get data (2x256 bytes)
	inir
	
	ld de,cf_sect
	ld hl,cf_need
	ld bc,#4
	ldir				; Copy sector #

	ei
	and a				; CF <- 0
	ret

lb_error
	ei
	scf
	ld e,#0x08			; Read error
	ret

	;
	; Write the buffer currently in the CF sector buffer back to the CF card
	;
write_buffer
	ld a,#1
	out (#2),a			; Sector count
	ld a,(iy+4)
	out (#3),a
	ld a,(iy+5)
	out (#4),a
	ld a,(iy+6)
	out (#5),a
	ld a,(iy+7)
	out (#6),a
	ld a,#0x30			; Write sector
	di				; No interrupts that will cause CF card to be confused
	out (#7),a

	ld bc,#0x0000			; Port 0, 256 bytes
	ld hl,#512			; CF card buffer
	
wb_waitdrq
	in a,(#7)			; CF status
	bit 0,a
	jr nz,lb_error			; Error!
	bit 3,a				; Wait for DRQ
	jr z,wb_waitdrq

	otir				; Write data (2x256 bytes)
	otir

wb_bsy
	in a,(#24)			; Wait for BSY to get set and clear
	bit 2,a
	jr z,lb_error			; CF card missing!
	cp #7
	jr nz,wb_bsy
	
	ei
	in a,(#7)
	rrca
	jr c,lb_error			; Error!
	and a				; CF <- 0
	ret

;
; NMI handing -- we receive an NMI when the CF card is removed.  Allow
; a DMA operation in process to complete, then report CMD/ERROR/READY
; with the error code of DRIVE NOT READY (80h) until the card is reinserted.
;
; We do a full initialization after a card has been inserted; this includes
; invalidating the CF sector buffer!
;
nmi
nmi_wait_dma
	in a,(#25)			; DMA status
	rrca
	jr c,nmi_wait_dma

	ld a,#0x80			; DRIVE NOT READY
	out (#9),a			; Auxilliary status
	ld a,#0x89			; CMD ERROR READY
	out (#8),a			; Primary status

nmi_wait_card	
	in a,(#24)			; CF status
	bit 2,a
	jp nz,(reset)			; If card inserted, start over...
	jr nmi_wait_card
	
sysinfo					; IY points here
command
	byte 0, 0, 0, 0			; Four byte command buffer
cf_sect
	byte 0, 0, 0, 0			; LBA of CF sector in CF buffer
cf_need
	byte 0, 0, 0, 0			; CF sector we need (temp for load_buffer)
abc_buf
	word 0				; Controller buffer addressed by command

sysinfo_end
